import abc
import copy
import datetime
import logging
from typing import Any

from openpyxl.styles import numbers
from openpyxl.worksheet import worksheet

from script_components import helpers


class BaseExcelTabProcessor(abc.ABC):
    DATE_FORMAT = "%Y-%m-%d"
    DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S"

    PYTHON_TYPE_TO_EXCEL_TYPE = {
        "str": numbers.FORMAT_TEXT,
        "int": numbers.FORMAT_NUMBER,
        "float": numbers.FORMAT_NUMBER_COMMA_SEPARATED1,
    }

    def __init__(self, data: list[Any], tab: worksheet.Worksheet):
        self.data = data
        self.tab = tab

    @abc.abstractmethod
    def run_tab_processing(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def clear_data_from_tab(self) -> None:
        raise NotImplementedError

    def _determine_data_type_and_value(self, value) -> dict:
        data_type = None

        match value:
            case bool():
                value = int(value)
                data_type = "int"
            case float():
                if value.is_integer():
                    value = int(value)
                    data_type = "int"
                else:
                    value = helpers.as_text(value)
                    data_type = "float"
            case datetime.datetime():
                value = value.strftime(self.DATE_TIME_FORMAT)
                data_type = "str"
            case datetime.date():
                value = value.strftime(self.DATE_FORMAT)
                data_type = "str"

        if data_type is None:
            data_type = "str"

        return {"data_type": data_type, "value": value}

    def _place_value_at_location(self, cell_location: str, row_data: dict) -> None:
        """
        Places the value in the Excel tab at the specified cell location.
        Example:
            cell_location = "A1"
            row_data = {"data_type": "int", "value": "123"}
        """
        formatted_row_value = helpers.as_text(row_data["value"])
        cell = self.tab[cell_location]
        cell.value = formatted_row_value

        if row_data["data_type"] in ["int", "float"]:
            number_data_format = "n"
            cell.data_type = number_data_format

        cell.number_format = self.PYTHON_TYPE_TO_EXCEL_TYPE[row_data["data_type"]]


class SingleColumnExcelTabProcessor(BaseExcelTabProcessor):
    VALUE_COLUMN = "B"
    FIRST_ROW_WITH_VALUE_TO_FILL = 1

    def __init__(self, data: list[Any], tab: worksheet.Worksheet):
        super().__init__(data, tab)
        if not data:
            raise ValueError("The data is missing for the requested site.")

        self.data_tuple = tuple(data[0])

    def run_tab_processing(self) -> None:
        self._fill_excel_column_with_values()

    def _fill_excel_column_with_values(self) -> None:
        excel_row_index = copy.copy(self.FIRST_ROW_WITH_VALUE_TO_FILL)

        for row_datum in self.data_tuple:
            transformed_row_data = self._determine_data_type_and_value(row_datum)
            cell_location = f"{self.VALUE_COLUMN}{excel_row_index}"
            self._place_value_at_location(cell_location, transformed_row_data)

            excel_row_index += 1

    def clear_data_from_tab(self) -> None:
        logging.warning(
            "Clearing data for the single-column values tabs not implemented. "
            "If you can see this, the implementation should be added!"
        )


class MultipleRowsExcelTabProcessor(BaseExcelTabProcessor):
    FIRST_ROW_TO_FILL = 3

    def run_tab_processing(self) -> None:
        self._fill_excel_rows_with_values()

    def _fill_excel_rows_with_values(self) -> None:
        excel_row_index = copy.copy(self.FIRST_ROW_TO_FILL)

        for row_data in self.data:
            excel_column_names_iterator = helpers.get_excel_column_names_iterator()
            for cell_datum in row_data:
                column_index = next(excel_column_names_iterator)
                cell_location = f"{column_index}{excel_row_index}"

                transformed_cell_datum = self._determine_data_type_and_value(cell_datum)
                self._place_value_at_location(cell_location, transformed_cell_datum)
            excel_row_index += 1

    def clear_data_from_tab(self) -> None:
        for row in self.tab.iter_rows(min_row=self.FIRST_ROW_TO_FILL):
            for cell in row:
                cell.value = None



from jinja2 import Environment, FileSystemLoader, select_autoescape, StrictUndefined
from pyprediktorutilities.shared import validate_folder, validate_file
import logging
import os

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Templating:
    """Simple wrapper around a templating engine, allowing for easy rendering of templates
    such as XMLs and HTMLs. Build on top of Jinja2
    
    Args:
        path (str): The path to the folder containing the templates
    
    Attributes:
        path (str): The path to the folder containing the templates
        env (jinja2.Environment): The environment to use for rendering
    
    Raises:
        FileNotFoundError: If the folder does not exist
        FileNotFoundError: If the template does not exist
        Exception: If the template could not be rendered
    
    Examples:
        >>> from pyprediktorutilities.templating import Templating
        >>> templating = Templating('templates')
        >>> templates = templating.list_templates()
        >>> template = templating.load_template('base.html')
        >>> rendered_template = templating.render(template, title='Home')
        >>> templating.render_to_file(template, 'output.html', title='Home')
    """
    
    def __init__(self, path: str) -> object:
        """Class initialization

        Args:
            path (str): The path to the folder containing the templates

        Raises:
            FileNotFoundError: If the folder does not exist

        Returns:
            object: Templating object
        """
        self.path = path
        # Check if folder exists
        try:
            validate_folder(path)
        except FileNotFoundError:
            errormsg = f"Folder {path} does not exist"
            logging.error(errormsg)
            raise FileNotFoundError(errormsg)
        
        # Establish the environment
        self.env = Environment(
            loader=FileSystemLoader(path), autoescape=select_autoescape(), undefined=StrictUndefined
        )
        
    def list_templates(self) -> list[str]:
        """Returns a list of templates in the folder

        Returns:
            list[str]: A list of strings containing the template names
        """
        return self.env.list_templates()

    def load_template(self, template: str) -> object:
        """Loads a template from the folder

        Args:
            template (str): The file name of the template to load

        Raises:
            FileNotFoundError: If the template does not exist

        Returns:
            object: The template object
        """
        # Check if template exists
        try:
            validate_file(os.path.join(self.path, template))
        except FileNotFoundError:
            errormsg = f"Template {template} not found in {self.path}"
            logging.error(errormsg)
            raise FileNotFoundError(errormsg)

        return self.env.get_template(template)

    def render(self, template: str, **kwargs) -> str:
        """Render a template with the given arguments and return a string

        Args:
            template (str): The file name of the template to load
            **kwargs: The arguments to pass to the template
            
        Raises:
            Exception: If the template could not be rendered

        Returns:
            str: The rendered template
        """
        template = self.env.get_template(template)
        try:
            return template.render(**kwargs)
        except Exception as e:
            errormsg = f"Could not render {template}: {e}"
            logging.error(errormsg)
            raise Exception(errormsg)
    
    def render_to_file(self, template: str, file: str, **kwargs) -> None:
        """Render a template with the given arguments and write to a file

        Args:
            template (str): The file name of the template to load
            file (str): The path and file name of the file to write to
            **kwargs: The arguments to pass to the template
        
        Raises:
            Exception: If the template could not be rendered
            FileNotFoundError: If the template does not exist
        
        Returns:
            None: None            
        """
        with open(file, "w") as f:
            f.write(self.render(template, **kwargs))



from unittest import mock

import pytest

import helpers
from pyprediktorutilities.dwh import dwh


@pytest.fixture()
@mock.patch(
    "pyprediktorutilities.dwh.dwh.Dwh._Dwh__get_list_of_available_and_supported_pyodbc_drivers",
    return_value={"available": ["Driver1"], "supported": ["Driver1"]},
)
def dwh_instance(_, monkeypatch):
    dwh_instance = dwh.Dwh(helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs())
    return dwh_instance


@pytest.fixture
def mock_pyodbc_connect(monkeypatch):
    mock_connection = mock.Mock()
    mock_cursor = mock.Mock()
    mock_connection.cursor.return_value = mock_cursor
    monkeypatch.setattr("pyodbc.connect", lambda *args, **kwargs: mock_connection)
    return mock_cursor



from pydantic import validate_call, EmailStr
import logging
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email import encoders
import os
from pyprediktorutilities.shared import validate_file

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class SendEmail:
    """Helper function to send emails with attachments using SMTP
    
    Args:
        server (str): SMTP server address
        port (int): SMTP server port
        username (str): SMTP server username
        password (str): SMTP server password
        
    Returns:
        Object: SendEmail object
    """
    
    @validate_call
    def __init__(self, server: str, port: int, username: str, password: str) -> object:
        """Class initialiser

        Args:
            server (str): SMTP server address
            port (int): SMTP server port
            username (str): SMTP server username
            password (str): SMTP server password

        Returns:
            object: SendEmail object
        """
        self.smtp_server = server
        self.smtp_server_port = port
        self.smtp_server_username = username
        self.smtp_server_password = password

    @validate_call
    def send_email(self, from_email: EmailStr, recipients: list[EmailStr], subject: str, body: str, files: list = []):
        """_summary_

        Args:
            from_email (email address): The sender email address
            recipients (list of email addresses): list of one or more recipient email addresses
            subject (str): The email subject
            body (str): The email body
            files (list, optional): A list of paths to files to include. Defaults to [].
        """
        
        if len(recipients) == 0:
            logging.error("No recipients specified")
            raise ValueError("No recipients specified")
        
        msg = MIMEMultipart()
        msg["From"] = str(from_email)
        msg["Subject"] = subject
        msg["To"] = ', '.join(str(e) for e in recipients)
        msg.attach(MIMEText(body, "plain"))

        for file in files:
            validate_file(file)
            
            with open(file, "rb") as f:
                part = MIMEBase("application", "octet-stream")
                part.set_payload(f.read())
                encoders.encode_base64(part)
                part.add_header(
                    "Content-Disposition",
                    f"attachment; filename={os.path.basename(file)}",
                )
                msg.attach(part)


        text = msg.as_string()
        reciplist = [str(e) for e in recipients]
        try:
            with smtplib.SMTP(self.smtp_server, self.smtp_server_port) as server:
                server.starttls()
                server.login(self.smtp_server_username, self.smtp_server_password)
                server.sendmail(str(from_email), reciplist, text)
                logging.info(f"Email sent to {reciplist} with subject {subject} and {len(files)} attachments")
        except smtplib.SMTPException as e:
            logging.error(f"Error: unable to send email to {reciplist} with subject {subject} and {len(files)} attachments")
            logging.error(e)
            raise e


if __name__ == "__main__":
    pass



import os
import tempfile
from unittest import TestCase, mock
from pyprediktorutilities.templating import Templating


class TestTemplating(TestCase):
    def setUp(self):
        self.templates_dir = tempfile.TemporaryDirectory()
        self.templating = Templating(self.templates_dir.name)

    def tearDown(self):
        self.templates_dir.cleanup()

    def test_initilization(self):
        with self.assertRaises(FileNotFoundError):
            nonexist = Templating('nonexistent_folder')

    def test_list_templates(self):
        templates = self.templating.list_templates()
        self.assertEqual(templates, [])

        with open(os.path.join(self.templates_dir.name, 'template1.html'), 'w') as f:
            f.write('<html><body>{{ title }}</body></html>')

        with open(os.path.join(self.templates_dir.name, 'template2.xml'), 'w') as f:
            f.write('<root><title>{{ title }}</title></root>')

        templates = self.templating.list_templates()
        self.assertEqual(set(templates), {'template1.html', 'template2.xml'})

    def test_load_template(self):
        with open(os.path.join(self.templates_dir.name, 'template1.html'), 'w') as f:
            f.write('<html><body>{{ title }}</body></html>')

        template = self.templating.load_template('template1.html')
        self.assertIsNotNone(template)

        with self.assertRaises(FileNotFoundError):
            self.templating.load_template('nonexistent_template.html')

    def test_render(self):
        with open(os.path.join(self.templates_dir.name, 'template1.html'), 'w') as f:
            f.write('<html><body>{{ title }}</body></html>')

        template = self.templating.load_template('template1.html')
        rendered_template = self.templating.render(template, title='Home')
        self.assertEqual(rendered_template, '<html><body>Home</body></html>')

    def test_render_missing_arg(self):
        string = """
        <html><body>
        {% for i in does_not_exist -%}
        {{ i }}
        {% endfor -%}
        </body></html>
        """
        with open(os.path.join(self.templates_dir.name, 'template1.html'), 'w') as f:  
            f.write(string)

        template = self.templating.load_template('template1.html')
        with self.assertRaises(Exception):
            self.templating.render(template)

    def test_render_to_file(self):
        with open(os.path.join(self.templates_dir.name, 'template1.html'), 'w') as f:
            f.write('<html><body>{{ title }}</body></html>')

        template = self.templating.load_template('template1.html')

        with tempfile.NamedTemporaryFile(delete=False) as f:
            self.templating.render_to_file(template, f.name, title='Home')
            with open(f.name, 'r') as rendered_file:
                self.assertEqual(rendered_file.read(), '<html><body>Home</body></html>')



import pytest
import smtplib
import logging
from pydantic import ValidationError

from pyprediktorutilities.send_email import SendEmail

srv = "smtp.gmail.com"
port = 587
usr = "username"
pwd = "password"
sndr = "nobody@somedomain.com"
rcpts = [sndr]
subj = "Test subject"
body = "Test body"

def test_send_email_wrong_port():
    with pytest.raises(ValidationError):
        SendEmail(srv, "not_valid_port", usr, pwd)

def test_send_email_missing_port():
    with pytest.raises(ValidationError):
        SendEmail(srv, None, usr, pwd)

def test_send_email_missing_server():
    with pytest.raises(ValidationError):
        SendEmail(None, port, usr, pwd)

def test_send_email_missing_username():
    with pytest.raises(ValidationError):
        SendEmail(srv, port, None, pwd)

def test_send_email_missing_password():
    with pytest.raises(ValidationError):
        SendEmail(srv, port, usr, None)

def test_send_email_missing_from_email():
    with pytest.raises(ValidationError):
        SendEmail(srv, port, usr, pwd).send_email(None, rcpts, subj, body)
    
def test_send_email_missing_recipients():
    with pytest.raises(ValidationError):
        SendEmail(srv, port, usr, pwd).send_email(sndr, None, subj, body)
        
def test_send_email_empty_recipients():
    with pytest.raises(ValueError):
        SendEmail(srv, port, usr, pwd).send_email(sndr, [], subj, body)

def test_send_email_missing_subject():
    with pytest.raises(ValidationError):
        SendEmail(srv, port, usr, pwd).send_email(sndr, rcpts, None, body)

def test_send_email_missing_body():
    with pytest.raises(ValidationError):
        SendEmail(srv, port, usr, pwd).send_email(sndr, rcpts, subj, None)

def test_send_email_no_such_files():
    with pytest.raises(FileNotFoundError):
        SendEmail(srv, port, usr, pwd).send_email(sndr, rcpts, subj, body, ["/no/such/file.txt"])

def test_mocked_send_email_forced_fail(mocker):
    # mock the smtplib.SMTP object
    mock_SMTP = mocker.MagicMock(name="pyprediktorutilities.send_email.smtplib.SMTP")
    mocker.patch("pyprediktorutilities.send_email.smtplib.SMTP", new=mock_SMTP)
    mock_SMTP.side_effect = smtplib.SMTPException
    with pytest.raises(smtplib.SMTPException):
        SendEmail(srv, port, usr, pwd).send_email(sndr, rcpts, subj, body)



import pyodbc
import logging
import pandas as pd
from typing import List, Any
from pydantic import validate_call

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Dwh:
    """Access a PowerView Data Warehouse or other SQL databases.

    Args:
        url (str): The URL of the sql server
        database (str): The name of the database
        username (str): The username
        password (str): The password

    Attributes:
        connection (pyodbc.Connection): The connection object
        cursor (pyodbc.Cursor): The cursor object
    """

    @validate_call
    def __init__(
        self,
        url: str,
        database: str,
        username: str,
        password: str,
        driver_index: int = -1,
    ) -> None:
        """Class initializer.

        Args:
            url (str): The URL of the sql server
            database (str): The name of the database
            username (str): The username
            password (str): The password
        """
        self.url = url
        self.driver = ""
        self.cursor = None
        self.database = database
        self.username = username
        self.password = password
        self.connection = None

        self.connection_string_template = (
            f"UID={self.username};"
            + f"PWD={self.password};"
            + "DRIVER={};"
            + f"SERVER={self.url};"
            + f"DATABASE={self.database};"
            + "TrustServerCertificate=yes;"
        )
        self.__set_driver(driver_index)
        self.connection_string = self.connection_string_template.format(self.driver)

        self.connection_attempts = 3

    def __enter__(self):
        self.__connect()
        return self

    @validate_call
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.connection is not None:
            self.__disconnect()

    """
    Public
    """

    @validate_call
    def fetch(self, query: str, to_dataframe: bool = False) -> List[Any]:
        """Execute the SQL query to get results from DWH and return the data.

        Use that method for getting data. That means that if you use SELECT or
        you'd like to call a stored procedure that returns one or more sets
        of data, that is the correct method to use.

        Use that method to GET.

        Args:
            query (str): The SQL query to execute.
            to_dataframe (bool): If True, return the results as a list
                of DataFrames.

        Returns:
            List[Any]: The results of the query. If DWH returns multiple
                data sets, this method is going to return a list
                of result sets (lists). If DWH returns a single data set,
                the method is going to return a list representing the single
                result set.

                If to_dataframe is True, the data inside each data set
                is going to be in DataFrame format.
        """
        self.__connect()
        try:
            self.cursor.execute(query)

            data_sets = []
            while True:
                columns = [col[0] for col in self.cursor.description]
                data_set = [dict(zip(columns, row)) for row in self.cursor.fetchall()]

                if to_dataframe:
                    data_sets.append(pd.DataFrame(data_set, columns=columns))
                else:
                    data_sets.append(data_set)

                if not self.cursor.nextset():
                    break

            return data_sets if len(data_sets) > 1 else data_sets[0]
        finally:
            self.__disconnect()  # prevent from leaving open transactions in DWH

    @validate_call
    def execute(self, query: str, *args, **kwargs) -> List[Any]:
        """Execute the SQL query and return the results.

        For instance, if we create a new record in DWH by calling
        a stored procedure returning the id of the inserted element or
        in our query we use `SELECT SCOPE_IDENTITY() AS LastInsertedId;`,
        the DWH is going to return data after executing our write request.

        Please note that here we expect a single result set. Therefore DWH
        is obligated to return only one data set and also we're obligated to
        construct our query according to this requirement.

        Use that method to CREATE, UPDATE, DELETE or execute business logic.
        To NOT use for GET.

        Args:
            query (str): The SQL query to execute.
            *args: Variable length argument list to pass to cursor.execute().
            **kwargs: Arbitrary keyword arguments to pass to cursor.execute().

        Returns:
            List[Any]: The results of the query.
        """
        self.__connect()
        try:
            self.cursor.execute(query, *args, **kwargs)
            result = self.cursor.fetchall()
            self.__commit()
            return result
        except Exception as e:
            logging.error(f"Failed to execute query: {e}")
            return []
        finally:
            self.__disconnect()  # prevent from leaving open transactions in DWH

    """
    Private - Driver
    """

    @validate_call
    def __set_driver(self, driver_index: int) -> None:
        """Sets the driver for the database connection.

        Args:
            driver (int): Index of the driver in the list of available drivers. If the index is -1 or
                in general below 0, pyPrediktorMapClient is going to choose
                the driver for you.

        Raises:
            ValueError: If no valid driver is found.
        """
        drivers = self.__get_list_of_available_and_supported_pyodbc_drivers()
        available_drivers = drivers["available"]
        supported_drivers = drivers["supported"]

        if not supported_drivers:
            raise ValueError("No supported ODBC drivers found.")
        if not available_drivers:
            raise Exception("Connection to the database cannot be established.")

        if driver_index < 0:
            self.driver = available_drivers[0]
        elif driver_index >= len(available_drivers):
            raise ValueError(
                f"Driver index {driver_index} is out of range. Please use "
                f"the __get_list_of_available_and_supported_pyodbc_drivers() method "
                f"to list all available drivers."
            )
        else:
            self.driver = available_drivers[driver_index]

    @validate_call
    def __get_list_of_supported_pyodbc_drivers(self) -> List[Any]:
        return pyodbc.drivers()

    @validate_call
    def __get_list_of_available_and_supported_pyodbc_drivers(
        self,
    ) -> dict:
        available_drivers = []
        supported_drivers = self.__get_list_of_supported_pyodbc_drivers()

        for driver in supported_drivers:
            try:
                connection_string_with_assigned_driver = (
                    self.connection_string_template.format(driver)
                )
                pyodbc.connect(connection_string_with_assigned_driver, timeout=3)
                available_drivers.append(driver)
            except pyodbc.Error as err:
                logger.info(f"Driver {driver} could not connect: {err}")

        drivers = {"available": available_drivers, "supported": supported_drivers}
        return drivers

    """
    Private - Connector & Disconnector
    """

    @validate_call
    def __connect(self) -> None:
        """Establishes a connection to the database."""
        if self.connection:
            return

        logging.info("Initiating connection to the database...")

        attempt = 0
        while attempt < self.connection_attempts:
            try:
                self.connection = pyodbc.connect(self.connection_string)
                if self.connection:
                    self.cursor = self.connection.cursor()
                    logging.info(f"Connected to the database on attempt {attempt + 1}")
                    return
                else:
                    logging.info(f"Connection is None on attempt {attempt + 1}")
                    raise pyodbc.Error("Failed to connect to the database")

            # Exceptions once thrown there is no point attempting
            except pyodbc.ProgrammingError as err:
                logger.error(
                    f"Programming Error {err.args[0] if err.args else 'No code'}: {err.args[1] if len(err.args) > 1 else 'No message'}"
                )
                logger.warning(
                    "There seems to be a problem with your code. Please "
                    "check your code and try again."
                )
                raise

            except (
                pyodbc.DataError,
                pyodbc.IntegrityError,
                pyodbc.NotSupportedError,
            ) as err:
                logger.error(
                    f"{type(err).__name__} {err.args[0] if err.args else 'No code'}: {err.args[1] if len(err.args) > 1 else 'No message'}"
                )
                raise

            # Exceptions when thrown we can continue attempting
            except pyodbc.OperationalError as err:
                logger.error(
                    f"Operational Error: {err.args[0] if err.args else 'No code'}: {err.args[1] if len(err.args) > 1 else 'No message'}"
                )
                logger.warning(
                    "Pyodbc is having issues with the connection. This "
                    "could be due to the wrong driver being used. Please "
                    "check your driver with "
                    "the __get_list_of_available_and_supported_pyodbc_drivers() method "
                    "and try again."
                )
                attempt += 1
                if self.__are_connection_attempts_reached(attempt):
                    break

            except (pyodbc.DatabaseError, pyodbc.Error) as err:
                logger.error(
                    f"{type(err).__name__} {err.args[0] if err.args else 'No code'}: {err.args[1] if len(err.args) > 1 else 'No message'}"
                )
                attempt += 1
                if self.__are_connection_attempts_reached(attempt):
                    break

        if not self.connection:
            raise pyodbc.Error("Failed to connect to the database")

    @validate_call
    def __are_connection_attempts_reached(self, attempt) -> bool:
        if attempt != self.connection_attempts:
            logger.warning("Retrying connection...")
            return False

        logger.error(
            f"Failed to connect to the DataWarehouse after "
            f"{self.connection_attempts} attempts."
        )
        return True

    @validate_call
    def __disconnect(self) -> None:
        """Closes the connection to the database."""
        if self.connection:
            self.connection.close()

            self.cursor = None
            self.connection = None

    """
    Private - Low level database operations
    """

    @validate_call
    def __commit(self) -> None:
        """Commits any changes to the database."""
        self.connection.commit()



from unittest import mock

from helpers import grs
from pyprediktorutilities.dwh import dwh_singleton


@mock.patch("pyprediktorutilities.dwh.dwh.Dwh._Dwh__set_driver")
def test_dwh_singleton_can_be_created_only_once(_, monkeypatch):
    driver_index = 0

    with mock.patch("pyprediktorutilities.dwh.dwh.pyodbc.connect"):
        db = dwh_singleton.DwhSingleton(grs(), grs(), grs(), grs(), driver_index)
        db_instance_address = id(db)

        db_2 = dwh_singleton.DwhSingleton(grs(), grs(), grs(), grs(), driver_index)
        db_2_instance_address = id(db_2)

    assert db_instance_address == db_2_instance_address



import unittest
from unittest import mock
import pytest
from pydantic import ValidationError

from pyprediktorutilities.shared import request_from_api, validate_file

URL = "http://someserver.somedomain.com/v1/"
return_json = [
    {
        "Id": "6:0:1029",
        "DisplayName": "IPVBaseCalculate",
        "BrowseName": "IPVBaseCalculate",
        "Props": [],
        "Vars": [],
    }
]


# This method will be used by the mock to replace requests
def mocked_requests(*args, **kwargs):
    class MockResponse:
        def __init__(self, json_data, status_code):
            self.json_data = json_data
            self.status_code = status_code
            self.raise_for_status = mock.Mock(return_value=False)

        def json(self):
            return self.json_data

    if args[0] == f"{URL}something":
        return MockResponse(return_json, 200)

    return MockResponse(None, 404)


class SharedFunctionsTestCase(unittest.TestCase):
    def test_requests_with_malformed_url(self):
        with pytest.raises(ValidationError):
            request_from_api(rest_url="No_valid_url", method="GET", endpoint="/")

    def test_requests_with_unsupported_method(self):
        with pytest.raises(ValidationError):
            request_from_api(rest_url=URL, method="NO_SUCH_METHOD", endpoint="/")

    @mock.patch("requests.get", side_effect=mocked_requests)
    def test_request_from_api_method_get(self, mock_get):
        result = request_from_api(rest_url=URL, method="GET", endpoint="something")
        assert result == return_json

    @mock.patch("requests.post", side_effect=mocked_requests)
    def test_request_from_api_method_post(self, mock_get):
        result = request_from_api(
            rest_url=URL, method="POST", endpoint="something", data="test"
        )
        assert result == return_json

    def test_validate_file_with_non_existing_file(self):
        with pytest.raises(FileNotFoundError):
            validate_file(file="No_such_file")

    def test_validate_file_with_existing_file(self):
        validate_file(file="tests/test_shared.py")

if __name__ == "__main__":
    unittest.main()



import requests
from pydantic import AnyUrl, validate_call
from typing import Literal
import logging
from pathlib import Path

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

@validate_call
def validate_folder(folder: str):
    if not Path(folder).is_dir():
        errormsg = f"Folder {folder} does not exist"
        logging.error(errormsg)
        raise FileNotFoundError(errormsg)

@validate_call
def validate_file(file: str):
    if not Path(file).is_file():
        errormsg = f"File {file} does not exist"
        logging.error(errormsg)
        raise FileNotFoundError(errormsg)


@validate_call
def request_from_api(
    rest_url: AnyUrl,
    method: Literal["GET", "POST"],
    endpoint: str,
    data: str = None,
    params: dict = None,
    headers: dict = None,
    extended_timeout: bool = False,
) -> str:
    """Function to perform request against a REST API

    Args:
        rest_url (str): The URL with trailing shash
        method (str): "GET" or "POST"
        endpoint (str): The last part of the url (without the leading slash)
        data (str): defaults to None but can contain the data to send to the endpoint
        headers (str): default to None but can contain the headers og the request
    Returns:
        JSON: The result if successfull
    """
    request_timeout = (3, 300 if extended_timeout else 27)
    combined_url = f"{rest_url}{endpoint}"
    if method == "GET":
        result = requests.get(combined_url, timeout=request_timeout, params=params, headers=headers)

    if method == "POST":
        result = requests.post(
            combined_url, data=data, headers=headers, timeout=request_timeout, params=params
        )

    result.raise_for_status()
    return result.json()



import argparse
import logging
from typing import Any

import openpyxl
from openpyxl import workbook

from script_components import excel_sheet_dwh
from script_components.tab_processors import (
    basic as basic_processors,
    specified as specified_processors,
)

logging.basicConfig(level=logging.INFO)


class DwhExcelSheetsFiller:
    TEMPLATE_NAME = "template.xlsx"
    TAB_NAME_TO_PROCESSOR_MAPPING = {
        "Main plant parameters": specified_processors.MainPlantParametersExcelTabProcessor,
        "DWH Extract Parameters": specified_processors.DWHExtractParametersExcelTabProcessor,
        "UA model parameters": specified_processors.UAModelParametersExcelTabProcessor,
        "SolarGIS parameters": specified_processors.SolarGISParametersExcelTabProcessor,
        "Sub facilities": basic_processors.MultipleRowsExcelTabProcessor,
        "String combiner parameters": basic_processors.MultipleRowsExcelTabProcessor,
        "Transformer parameters": basic_processors.MultipleRowsExcelTabProcessor,
        "Module-Mounting stru parameters": basic_processors.MultipleRowsExcelTabProcessor,
        "Inverter parameters": basic_processors.MultipleRowsExcelTabProcessor,
        "String Channel parameters": basic_processors.MultipleRowsExcelTabProcessor,
        "Tracker parameters": basic_processors.MultipleRowsExcelTabProcessor,
        "Weather station parameters": basic_processors.MultipleRowsExcelTabProcessor,
        "Sub Station": basic_processors.MultipleRowsExcelTabProcessor,
        "Sub Station Transformer": basic_processors.MultipleRowsExcelTabProcessor,
        "Feeder": basic_processors.MultipleRowsExcelTabProcessor,
        "Grid Connection Point": basic_processors.MultipleRowsExcelTabProcessor,
        "PPC Parameters": basic_processors.MultipleRowsExcelTabProcessor,
        "Meter parameters": basic_processors.MultipleRowsExcelTabProcessor,
    }

    def __init__(self, dwh: excel_sheet_dwh.ExcelSheetDwh, site: str):
        self.dwh = dwh
        self.site = site

    def run(self, output_file_name: str) -> None:
        try:
            excel_file = openpyxl.load_workbook(self.TEMPLATE_NAME)
        except FileNotFoundError:
            raise FileNotFoundError(
                f"The template file has not been found. "
                f"Paste the file named {self.TEMPLATE_NAME} in the same directory as the script."
            )

        self._fill_excel_file(excel_file)

        self._save_excel_sheet(excel_file, output_file_name)
        excel_file.close()

    def _fill_excel_file(self, excel_file: workbook.Workbook) -> None:
        for tab_name in excel_file.sheetnames:
            if tab_name in self.TAB_NAME_TO_PROCESSOR_MAPPING:
                logging.info(f"Checking if the tab named {tab_name!r} needs data...")
                self._fill_tab_with_data(excel_file, tab_name)
                logging.info(f"The tab named {tab_name!r} has been filled with data.")
            else:
                logging.info(f"The tab named {tab_name!r} has been skipped.")

    def _fill_tab_with_data(self, excel_file: workbook.Workbook, tab_name: str) -> None:
        data_to_write = self.dwh.get_data_for_tab(tab_name=tab_name, site=self.site)

        row_count = len(data_to_write)
        logging.info(f"Number of rows to write for {tab_name!r}: {row_count}")

        tab_processor = self.__select_tab_processor(data_to_write, excel_file, tab_name)

        if row_count == 0:
            tab_processor.clear_data_from_tab()
        else:
            tab_processor.run_tab_processing()

    def __select_tab_processor(
        self, data_to_write: list[Any], excel_file: workbook.Workbook, tab_name: str
    ) -> basic_processors.BaseExcelTabProcessor:
        tab_processor_class = self.TAB_NAME_TO_PROCESSOR_MAPPING[tab_name]
        tab_processor = tab_processor_class(data_to_write, excel_file[tab_name])
        return tab_processor

    def _save_excel_sheet(
        self, excel_sheet: workbook.Workbook, output_file_name: str
    ) -> None:
        logging.info(f"Saving the Excel sheet to {output_file_name!r}...")
        excel_sheet.save(output_file_name)
        logging.info("Saving the Excel sheet has been completed!")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--db_url",
        type=str,
        default="localhost",
        help="The URL of the MSSQL server",
    )
    parser.add_argument(
        "--db_name",
        type=str,
        default="master",
        help="The name of the database",
    )
    parser.add_argument(
        "--db_username",
        type=str,
        default="sa",
        help="The username",
    )
    parser.add_argument(
        "--db_password",
        type=str,
        default="Password123",
        help="The password",
    )
    parser.add_argument(
        "--output_file",
        type=str,
        default="output.xlsx",
        help="The name of the output file",
    )
    parser.add_argument(
        "--site",
        type=str,
        default="None",
        help="Site to fetch data for",
    )
    args = parser.parse_args()
    output_file = args.output_file
    db_url = args.db_url
    db_name = args.db_name
    db_username = args.db_username
    db_password = args.db_password
    site = args.site

    data_warehouse = excel_sheet_dwh.ExcelSheetDwh(
        url=db_url,
        database=db_name,
        username=db_username,
        password=db_password,
    )
    DwhExcelSheetsFiller(dwh=data_warehouse, site=site).run(
        output_file_name=output_file
    )



import pytest
import logging
from pydantic import ValidationError

from pyprediktorutilities.file_transfer import SFTPClient

server = "someserver.somedomain.com"
username = "username"
password = "password"
port = 22

def test_instance_of_sftp_client():
    sftp_client = SFTPClient(server=server, username=username, password=password, port=port)
    assert isinstance(sftp_client, SFTPClient)
    assert sftp_client.server == server
    assert sftp_client.username == username
    assert sftp_client.password == password
    assert sftp_client.port == port

def test_sftp_client_with_invalid_server():
    with pytest.raises(ValidationError):
        SFTPClient(server=None, username=username, password=password, port=port)

def test_sftp_client_with_invalid_port():
    with pytest.raises(ValidationError):
        SFTPClient(server=server, username=username, password=password, port="No_port")

def test_sftp_client_with_invalid_username():
    with pytest.raises(ValidationError):
        SFTPClient(server=server, username=123, password=password, port=port)

def test_sftp_client_with_invalid_password():
    with pytest.raises(ValidationError):
        SFTPClient(server=server, username=username, password=None, port=port)

def test_sftp_client_with_missing_file():
    with pytest.raises(FileNotFoundError):
        SFTPClient(server=server, username=username, password=password, port=port).upload(files=[], remote_directory="/tmp")

def test_sftp_client_with_nonexisting_file():
    with pytest.raises(FileNotFoundError):
        SFTPClient(server=server, username=username, password=password, port=port).upload(files=["No_such_file"], remote_directory="/tmp")


import copy
from typing import Any

from openpyxl.worksheet import worksheet

from script_components.tab_processors import basic as basic_processors


class MainPlantParametersExcelTabProcessor(
    basic_processors.SingleColumnExcelTabProcessor
):
    FIRST_ROW_WITH_VALUE_TO_FILL = 4

    TARGET_DWH_OUTPUT_LENGTH = 135
    INDEX_PRIOR_TO_MISSING_GRID_STATION_ID = 14

    def __init__(self, data: list[Any], tab: worksheet.Worksheet):
        super().__init__(data, tab)
        self.is_missing_grid_station_id = False

    def _fill_excel_column_with_values(self) -> None:
        self.is_missing_grid_station_id = (
            len(self.data_tuple) == self.TARGET_DWH_OUTPUT_LENGTH
        )
        excel_row_index = copy.copy(self.FIRST_ROW_WITH_VALUE_TO_FILL)

        for row_datum in self.data_tuple:
            transformed_row_data = self._determine_data_type_and_value(row_datum)
            cell_location = f"{self.VALUE_COLUMN}{excel_row_index}"
            self._place_value_at_location(cell_location, transformed_row_data)

            if self.is_missing_grid_station_id:
                if excel_row_index == self.INDEX_PRIOR_TO_MISSING_GRID_STATION_ID:
                    excel_row_index += 1
                    self.tab[f"{self.VALUE_COLUMN}{excel_row_index}"] = ""

            excel_row_index += 1


class DWHExtractParametersExcelTabProcessor(
    basic_processors.SingleColumnExcelTabProcessor
):
    FIRST_ROW_WITH_VALUE_TO_FILL = 3


class UAModelParametersExcelTabProcessor(
    basic_processors.SingleColumnExcelTabProcessor
):
    FIRST_ROW_WITH_VALUE_TO_FILL = 3


class SolarGISParametersExcelTabProcessor(
    basic_processors.SingleColumnExcelTabProcessor
):
    FIRST_ROW_WITH_VALUE_TO_FILL = 5



from pydantic import validate_call
import os
import logging
import paramiko
from pyprediktorutilities.shared import validate_file

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class SFTPClient:
    """An SFTP client to upload files to a remote server
    
    Args:
        server (AnyUrl): The server address
        username (str): The username
        password (str): The password
        port (int): The port number
        
    Returns:
        Object: SFTPClient object
    """
    @validate_call
    def __init__(self, server: str, username: str, password: str, port: int) -> object:
        self.server = server
        self.username = username
        self.password = password
        self.port = port

    @validate_call
    def upload(self, files: list[str], remote_directory: str):
        """Upload the files to the remote directory, create the directory
        recursively if it does not exist (and the remote server allows it)

        Args:
            files (list[str]): _description_
            remote_directory (str): _description_
        """
        
        if not files:
            raise FileNotFoundError("No files to upload")
        
        for file in files:
            validate_file(file)
        
        with paramiko.SSHClient() as ssh:
            ssh.load_system_host_keys()
            ssh.connect(
                self.server,
                port=self.port,
                username=self.username,
                password=self.password,
            )

            with ssh.open_sftp() as sftp:
                # Check if remote directory exists or create it
                try:
                    sftp.stat(remote_directory)
                except FileNotFoundError:
                    sftp.mkdir(remote_directory)

                # Upload each file
                for file in files:
                    remote_path = f"{remote_directory}/{os.path.basename(file)}"
                    sftp.put(file, remote_path)
                    logging.info(
                        f"Uploaded {file} to {remote_path} successfully!"
                    )




import logging
from typing import Any

from pyprediktorutilities.dwh.dwh_singleton import DwhSingleton

logging.basicConfig(level=logging.INFO)


QUERIES_LIST = {
    "Main plant parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='Main plant parameters';",
    "DWH Extract Parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='DWH Extract Parameters';",
    "UA model parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='UA model parameters';",
    "SolarGIS parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='SolarGIS parameters';",
    "Sub facilities": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='Sub facilities';",
    "String combiner parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='String combiner';",
    "Transformer parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='Transformer parameters';",
    "Module-Mounting stru parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='Module-Mounting stru parameters';",
    "Inverter parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='Inverter parameters';",
    "String Channel parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='String Channel parameters';",
    "Tracker parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='Tracker parameters';",
    "Weather station parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='Weather station parameters';",
    "Sub Station": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='Sub Station';",
    "Sub Station Transformer": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='Sub Station Transformer';",
    "Feeder": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='Feeder';",
    "Grid Connection Point": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='Grid Connection Point';",
    "PPC Parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='PPC Parameters';",
    "Meter parameters": "exec [dwetl].[GetPlantParametersData] @Facilityname='{}', @DataType='Meter parameters';",
}


class ExcelSheetDwh(DwhSingleton):
    def get_data_for_tab(self, tab_name: str, site: str) -> list[Any]:
        query = QUERIES_LIST[tab_name].format(site)
        result = self.execute(query=query)

        if not result:
            logging.warning(
                f"No data fetched for tab: {tab_name!r} for site: {site!r}."
            )
        return result



import itertools
import string

from typing import Iterator


def as_text(value) -> str:
    if value is None:
        return ""
    return str(value)


def get_excel_column_names_iterator() -> Iterator:
    single_letters = list(string.ascii_uppercase)
    double_letters = []
    for letter in single_letters:
        uppercase_aa_zz_list = [
            f"{letter}{second_letter}" for second_letter in single_letters
        ]
        double_letters = [*double_letters, *uppercase_aa_zz_list]
    return itertools.chain(single_letters, double_letters)



import helpers
from pyprediktorutilities.dwh import dwh

from unittest import mock

import pandas as pd
import pyodbc
import pytest
from pandas.testing import assert_frame_equal


class TestDwh:
    @mock.patch("pyprediktorutilities.dwh.dwh.Dwh._Dwh__set_driver")
    def test_init_when_instantiate_db_then_instance_is_created(self, _, monkeypatch):
        driver_index = 0

        with mock.patch("pyprediktorutilities.dwh.dwh.pyodbc.connect"):
            db = dwh.Dwh(
                helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs(), driver_index
            )

        assert db is not None

    @mock.patch("pyprediktorutilities.dwh.dwh.Dwh._Dwh__set_driver")
    def test_dwh_connect_when_connection_is_present(self, _, dwh_instance):
        dwh_instance.connection = mock.Mock()
        result = dwh_instance._Dwh__connect()
        assert result is None

    @mock.patch("pyprediktorutilities.dwh.dwh.Dwh._Dwh__set_driver")
    @mock.patch("pyprediktorutilities.dwh.dwh.Dwh._Dwh__connect")
    @mock.patch("pyprediktorutilities.dwh.dwh.Dwh._Dwh__disconnect")
    def test_dwh_use_as_context_manager_when_connection_is_present(
        self, disconnect_mock, connect_mock, set_driver_mock, dwh_instance
    ):
        with dwh_instance:
            pass

        assert not disconnect_mock.called

    def test_init_when_instantiate_dwh_but_driver_index_is_not_passed_then_instance_is_created(
        self,
        monkeypatch,
    ):
        # Mock the connection method to return a mock connection with a mock cursor
        mock_cursor = mock.Mock()
        mock_connection = mock.Mock()
        mock_connection.cursor.return_value = mock_cursor
        monkeypatch.setattr("pyodbc.connect", lambda *args, **kwargs: mock_connection)
        monkeypatch.setattr("pyodbc.drivers", lambda: ["Driver1", "Driver2"])

        db = dwh.Dwh(helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs())
        assert db is not None
        assert db.driver == "Driver1"

    """
    fetch
    """

    def test_fetch_when_to_dataframe_is_false_and_no_data_is_returned_then_return_empty_list(
        self,
        monkeypatch,
    ):
        query = "SELECT * FROM mytable"
        driver_index = 2

        expected_result = []

        # Mock the cursor's fetchall methods
        mock_cursor = mock.Mock()
        mock_cursor.fetchall.return_value = []
        mock_cursor.nextset.return_value = False
        mock_cursor.description = [
            ("plantname", None),
            ("resource_id", None),
            ("api_key", None),
            ("ExtForecastTypeKey", None),
            ("hours", None),
            ("output_parameters", None),
            ("period", None),
        ]

        # Mock the connection method to return a mock connection with a mock cursor
        mock_connection = mock.Mock()
        mock_connection.cursor.return_value = mock_cursor
        monkeypatch.setattr("pyodbc.connect", lambda *args, **kwargs: mock_connection)
        monkeypatch.setattr("pyodbc.drivers", lambda: ["Driver1", "Driver2", "Driver3"])

        db = dwh.Dwh(
            helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs(), driver_index
        )
        actual_result = db.fetch(query)

        mock_cursor.execute.assert_called_once_with(query)
        assert actual_result == expected_result
        assert db.connection is None

    def test_fetch_when_to_dataframe_is_false_and_single_data_set_is_returned_then_return_list_representing_single_data_set(
        self,
        monkeypatch,
    ):
        query = "SELECT * FROM mytable"
        driver_index = 2
        data_returned_by_db = [
            (
                "XY-ZK",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                13,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
            (
                "XY-ZK",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                14,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
            (
                "KL-MN",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                13,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
        ]

        expected_result = [
            {
                "plantname": "XY-ZK",
                "resource_id": "1234-abcd-efgh-5678",
                "api_key": "SOME_KEY",
                "ExtForecastTypeKey": 13,
                "hours": 168,
                "output_parameters": "pv_power_advanced",
                "period": "PT15M",
            },
            {
                "plantname": "XY-ZK",
                "resource_id": "1234-abcd-efgh-5678",
                "api_key": "SOME_KEY",
                "ExtForecastTypeKey": 14,
                "hours": 168,
                "output_parameters": "pv_power_advanced",
                "period": "PT15M",
            },
            {
                "plantname": "KL-MN",
                "resource_id": "1234-abcd-efgh-5678",
                "api_key": "SOME_KEY",
                "ExtForecastTypeKey": 13,
                "hours": 168,
                "output_parameters": "pv_power_advanced",
                "period": "PT15M",
            },
        ]

        # Mock the cursor's fetchall methods
        mock_cursor = mock.Mock()
        mock_cursor.fetchall.return_value = data_returned_by_db
        mock_cursor.nextset.return_value = False
        mock_cursor.description = [
            ("plantname", None),
            ("resource_id", None),
            ("api_key", None),
            ("ExtForecastTypeKey", None),
            ("hours", None),
            ("output_parameters", None),
            ("period", None),
        ]

        # Mock the connection method to return a mock connection with a mock cursor
        mock_connection = mock.Mock()
        mock_connection.cursor.return_value = mock_cursor
        monkeypatch.setattr("pyodbc.connect", lambda *args, **kwargs: mock_connection)
        monkeypatch.setattr("pyodbc.drivers", lambda: ["Driver1", "Driver2", "Driver3"])

        db = dwh.Dwh(
            helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs(), driver_index
        )
        actual_result = db.fetch(query)

        mock_cursor.execute.assert_called_once_with(query)
        assert actual_result == expected_result
        assert db.connection is None

    def test_fetch_when_to_dataframe_is_false_and_multiple_data_sets_are_returned_then_return_list_of_lists_representing_multiple_data_sets(
        self,
        monkeypatch,
    ):
        query = "SELECT * FROM mytable"
        driver_index = 2
        data_returned_by_db_set_one = [
            (
                "XY-ZK",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                13,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
            (
                "XY-ZK",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                14,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
            (
                "KL-MN",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                13,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
        ]
        data_returned_by_db_set_two = [
            (
                "ALPHA",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                13,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
            (
                "BETA",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                14,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
        ]

        expected_result = [
            [
                {
                    "plantname": "XY-ZK",
                    "resource_id": "1234-abcd-efgh-5678",
                    "api_key": "SOME_KEY",
                    "ExtForecastTypeKey": 13,
                    "hours": 168,
                    "output_parameters": "pv_power_advanced",
                    "period": "PT15M",
                },
                {
                    "plantname": "XY-ZK",
                    "resource_id": "1234-abcd-efgh-5678",
                    "api_key": "SOME_KEY",
                    "ExtForecastTypeKey": 14,
                    "hours": 168,
                    "output_parameters": "pv_power_advanced",
                    "period": "PT15M",
                },
                {
                    "plantname": "KL-MN",
                    "resource_id": "1234-abcd-efgh-5678",
                    "api_key": "SOME_KEY",
                    "ExtForecastTypeKey": 13,
                    "hours": 168,
                    "output_parameters": "pv_power_advanced",
                    "period": "PT15M",
                },
            ],
            [
                {
                    "plantname": "ALPHA",
                    "resource_id": "1234-abcd-efgh-5678",
                    "api_key": "SOME_KEY",
                    "ExtForecastTypeKey": 13,
                    "hours": 168,
                    "output_parameters": "pv_power_advanced",
                    "period": "PT15M",
                },
                {
                    "plantname": "BETA",
                    "resource_id": "1234-abcd-efgh-5678",
                    "api_key": "SOME_KEY",
                    "ExtForecastTypeKey": 14,
                    "hours": 168,
                    "output_parameters": "pv_power_advanced",
                    "period": "PT15M",
                },
            ],
        ]

        # Mock the cursor's fetchall methods
        mock_cursor = mock.Mock()
        mock_cursor.fetchall.side_effect = [
            data_returned_by_db_set_one,
            data_returned_by_db_set_two,
        ]
        mock_cursor.nextset.side_effect = [True, False]
        mock_cursor.description = [
            ("plantname", None),
            ("resource_id", None),
            ("api_key", None),
            ("ExtForecastTypeKey", None),
            ("hours", None),
            ("output_parameters", None),
            ("period", None),
        ]

        # Mock the connection method to return a mock connection with a mock cursor
        mock_connection = mock.Mock()
        mock_connection.cursor.return_value = mock_cursor
        monkeypatch.setattr("pyodbc.connect", lambda *args, **kwargs: mock_connection)
        monkeypatch.setattr("pyodbc.drivers", lambda: ["Driver1", "Driver2", "Driver3"])

        db = dwh.Dwh(
            helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs(), driver_index
        )
        actual_result = db.fetch(query)

        mock_cursor.execute.assert_called_once_with(query)
        assert actual_result == expected_result
        assert db.connection is None

    def test_fetch_when_to_dataframe_is_true_and_no_data_is_returned_then_return_empty_dataframe(
        self,
        monkeypatch,
    ):
        query = "SELECT * FROM mytable"
        driver_index = 2

        # Mock the cursor's fetchall methods
        mock_cursor = mock.Mock()
        mock_cursor.fetchall.return_value = []
        mock_cursor.nextset.return_value = False
        mock_cursor.description = [
            ("plantname", None),
            ("resource_id", None),
            ("api_key", None),
            ("ExtForecastTypeKey", None),
            ("hours", None),
            ("output_parameters", None),
            ("period", None),
        ]

        # Mock the connection method to return a mock connection with a mock cursor
        mock_connection = mock.Mock()
        mock_connection.cursor.return_value = mock_cursor
        monkeypatch.setattr("pyodbc.connect", lambda *args, **kwargs: mock_connection)
        monkeypatch.setattr("pyodbc.drivers", lambda: ["Driver1", "Driver2", "Driver3"])

        db = dwh.Dwh(
            helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs(), driver_index
        )
        actual_result = db.fetch(query, True)

        mock_cursor.execute.assert_called_once_with(query)
        assert actual_result.empty
        assert db.connection is None

    def test_fetch_when_to_dataframe_is_true_and_single_data_set_is_returned_then_return_dataframe(
        self,
        monkeypatch,
    ):
        query = "SELECT * FROM mytable"
        driver_index = 2
        data_returned_by_db = [
            (
                "XY-ZK",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                13,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
            (
                "XY-ZK",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                14,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
            (
                "KL-MN",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                13,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
        ]

        expected_result = [
            {
                "plantname": "XY-ZK",
                "resource_id": "1234-abcd-efgh-5678",
                "api_key": "SOME_KEY",
                "ExtForecastTypeKey": 13,
                "hours": 168,
                "output_parameters": "pv_power_advanced",
                "period": "PT15M",
            },
            {
                "plantname": "XY-ZK",
                "resource_id": "1234-abcd-efgh-5678",
                "api_key": "SOME_KEY",
                "ExtForecastTypeKey": 14,
                "hours": 168,
                "output_parameters": "pv_power_advanced",
                "period": "PT15M",
            },
            {
                "plantname": "KL-MN",
                "resource_id": "1234-abcd-efgh-5678",
                "api_key": "SOME_KEY",
                "ExtForecastTypeKey": 13,
                "hours": 168,
                "output_parameters": "pv_power_advanced",
                "period": "PT15M",
            },
        ]
        expected_df = pd.DataFrame(expected_result)

        # Mock the cursor's fetchall methods
        mock_cursor = mock.Mock()
        mock_cursor.fetchall.return_value = data_returned_by_db
        mock_cursor.nextset.return_value = False
        mock_cursor.description = [
            ("plantname", None),
            ("resource_id", None),
            ("api_key", None),
            ("ExtForecastTypeKey", None),
            ("hours", None),
            ("output_parameters", None),
            ("period", None),
        ]

        # Mock the connection method to return a mock connection with a mock cursor
        mock_connection = mock.Mock()
        mock_connection.cursor.return_value = mock_cursor
        monkeypatch.setattr("pyodbc.connect", lambda *args, **kwargs: mock_connection)
        monkeypatch.setattr("pyodbc.drivers", lambda: ["Driver1", "Driver2", "Driver3"])

        db = dwh.Dwh(
            helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs(), driver_index
        )
        actual_result = db.fetch(query, True)

        mock_cursor.execute.assert_called_once_with(query)
        assert_frame_equal(
            actual_result.reset_index(drop=True),
            expected_df.reset_index(drop=True),
            check_dtype=False,
        )
        assert db.connection is None

    def test_fetch_when_to_dataframe_is_true_and_multiple_data_sets_are_returned_then_return_list_of_dataframes_representing_multiple_data_sets(
        self,
        monkeypatch,
    ):
        query = "SELECT * FROM mytable"
        driver_index = 2
        data_returned_by_db_set_one = [
            (
                "XY-ZK",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                13,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
            (
                "XY-ZK",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                14,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
            (
                "KL-MN",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                13,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
        ]
        data_returned_by_db_set_two = [
            (
                "ALPHA",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                13,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
            (
                "BETA",
                "1234-abcd-efgh-5678",
                "SOME_KEY",
                14,
                168,
                "pv_power_advanced",
                "PT15M",
            ),
        ]

        expected_result_set_one = [
            {
                "plantname": "XY-ZK",
                "resource_id": "1234-abcd-efgh-5678",
                "api_key": "SOME_KEY",
                "ExtForecastTypeKey": 13,
                "hours": 168,
                "output_parameters": "pv_power_advanced",
                "period": "PT15M",
            },
            {
                "plantname": "XY-ZK",
                "resource_id": "1234-abcd-efgh-5678",
                "api_key": "SOME_KEY",
                "ExtForecastTypeKey": 14,
                "hours": 168,
                "output_parameters": "pv_power_advanced",
                "period": "PT15M",
            },
            {
                "plantname": "KL-MN",
                "resource_id": "1234-abcd-efgh-5678",
                "api_key": "SOME_KEY",
                "ExtForecastTypeKey": 13,
                "hours": 168,
                "output_parameters": "pv_power_advanced",
                "period": "PT15M",
            },
        ]
        expected_result_set_two = [
            {
                "plantname": "ALPHA",
                "resource_id": "1234-abcd-efgh-5678",
                "api_key": "SOME_KEY",
                "ExtForecastTypeKey": 13,
                "hours": 168,
                "output_parameters": "pv_power_advanced",
                "period": "PT15M",
            },
            {
                "plantname": "BETA",
                "resource_id": "1234-abcd-efgh-5678",
                "api_key": "SOME_KEY",
                "ExtForecastTypeKey": 14,
                "hours": 168,
                "output_parameters": "pv_power_advanced",
                "period": "PT15M",
            },
        ]
        expected_df_set_one = pd.DataFrame(expected_result_set_one)
        expected_df_set_two = pd.DataFrame(expected_result_set_two)

        # Mock the cursor's fetchall methods
        mock_cursor = mock.Mock()
        mock_cursor.fetchall.side_effect = [
            data_returned_by_db_set_one,
            data_returned_by_db_set_two,
        ]
        mock_cursor.nextset.side_effect = [True, False]
        mock_cursor.description = [
            ("plantname", None),
            ("resource_id", None),
            ("api_key", None),
            ("ExtForecastTypeKey", None),
            ("hours", None),
            ("output_parameters", None),
            ("period", None),
        ]

        # Mock the connection method to return a mock connection with a mock cursor
        mock_connection = mock.Mock()
        mock_connection.cursor.return_value = mock_cursor
        monkeypatch.setattr("pyodbc.connect", lambda *args, **kwargs: mock_connection)
        monkeypatch.setattr("pyodbc.drivers", lambda: ["Driver1", "Driver2", "Driver3"])

        db = dwh.Dwh(
            helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs(), driver_index
        )
        actual_result = db.fetch(query, True)

        mock_cursor.execute.assert_called_once_with(query)
        assert_frame_equal(
            actual_result[0].reset_index(drop=True),
            expected_df_set_one,
            check_dtype=False,
        )
        assert_frame_equal(
            actual_result[1].reset_index(drop=True),
            expected_df_set_two,
            check_dtype=False,
        )
        assert db.connection is None

    """
    execute
    """

    @mock.patch("pyprediktorutilities.dwh.dwh.Dwh._Dwh__set_driver")
    def test_execute_when_parameter_passed_then_fetch_results_and_return_data(
        self, _, monkeypatch
    ):
        query = "INSERT INTO mytable VALUES (?, ?)"
        param_one = "John"
        param_two = "Smith"
        driver_index = 0
        expected_result = [{"id": 13}]

        # Mock the cursor and execute
        mock_cursor = mock.Mock()
        mock_execute = mock.Mock()

        # Mock the connection method to return a mock connection with a mock cursor
        mock_connection = mock.Mock()
        mock_connection.cursor.return_value = mock_cursor

        monkeypatch.setattr(
            "pyodbc.connect",
            mock.Mock(return_value=mock_connection),
        )

        # Mock the fetch method
        mock_fetch = mock.Mock(return_value=expected_result)
        mock_cursor.execute = mock_execute
        mock_cursor.fetchall = mock_fetch

        db = dwh.Dwh(
            helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs(), driver_index
        )
        actual_result = db.execute(query, param_one, param_two)

        mock_execute.assert_called_once_with(query, param_one, param_two)
        mock_fetch.assert_called_once()
        assert actual_result == expected_result
        assert db.connection is None

    def test_init_when_instantiate_db_but_no_pyodbc_drivers_available_then_throw_exception(
        self, monkeypatch
    ):
        driver_index = 0
        monkeypatch.setattr("pyodbc.drivers", lambda: ["DRIVER1"])
        monkeypatch.setattr(
            dwh.Dwh,
            "_Dwh__get_list_of_available_and_supported_pyodbc_drivers",
            lambda self: {"available": [], "supported": []},
        )

        with pytest.raises(ValueError, match="No supported ODBC drivers found."):
            dwh.Dwh(
                helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs(), driver_index
            )

    def test_init_with_out_of_range_driver_index(self, monkeypatch):
        driver_index = 1
        monkeypatch.setattr("pyodbc.drivers", lambda: ["DRIVER1"])
        monkeypatch.setattr(
            dwh.Dwh,
            "_Dwh__get_list_of_available_and_supported_pyodbc_drivers",
            lambda self: {"available": ["DRIVER1"], "supported": ["DRIVER1"]},
        )

        with pytest.raises(ValueError, match="Driver index 1 is out of range."):
            dwh.Dwh(
                helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs(), driver_index
            )

    def test_init_sets_driver_into_connection_string(self, monkeypatch):
        driver_name = "MY_MSSQL_DRIVER"
        monkeypatch.setattr("pyodbc.drivers", lambda: [driver_name])
        monkeypatch.setattr(
            dwh.Dwh,
            "_Dwh__get_list_of_available_and_supported_pyodbc_drivers",
            lambda self: {"available": [driver_name], "supported": [driver_name]},
        )

        dwh_instance = dwh.Dwh(
            helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs()
        )

        assert driver_name in dwh_instance.connection_string

    @mock.patch("pyodbc.connect")
    def test_context_manager_enter(self, _, dwh_instance):
        assert dwh_instance.__enter__() == dwh_instance

    @mock.patch("pyodbc.connect")
    def test_exit_without_connection(self, dwh_instance):
        dwh_instance.connection = None
        dwh_instance.__exit__(None, None, None)

    @mock.patch("pyodbc.connect")
    def test_exit_with_open_connection_and_cleanup(self, _, dwh_instance):
        mock_connection = mock.Mock()
        dwh_instance.connection = mock_connection
        dwh_instance.__exit__(None, None, None)

        mock_connection.close.assert_called_once()
        assert (
            dwh_instance.connection is None
        ), "Connection should be set to None after exit"

    @pytest.mark.parametrize("to_dataframe", [False, True])
    def test_fetch_multiple_datasets(
        self, mock_pyodbc_connect, dwh_instance, to_dataframe
    ):
        data1 = [("value1", 1), ("value2", 2)]
        data2 = [("value3", 3), ("value4", 4)]
        mock_pyodbc_connect.fetchall.side_effect = [data1, data2]
        mock_pyodbc_connect.nextset.side_effect = [True, False]
        mock_pyodbc_connect.description = [
            ("column1", None),
            ("column2", None),
        ]

        result = dwh_instance.fetch("SELECT * FROM test_table", to_dataframe)

        if to_dataframe:
            expected1 = pd.DataFrame(data1, columns=["column1", "column2"])
            expected2 = pd.DataFrame(data2, columns=["column1", "column2"])
            assert len(result) == 2
            assert_frame_equal(result[0], expected1)
            assert_frame_equal(result[1], expected2)
        else:
            expected = [
                [
                    {"column1": "value1", "column2": 1},
                    {"column1": "value2", "column2": 2},
                ],
                [
                    {"column1": "value3", "column2": 3},
                    {"column1": "value4", "column2": 4},
                ],
            ]
            assert result == expected

    def test_execute_with_fetch_error(self, dwh_instance, mock_pyodbc_connect, caplog):
        mock_pyodbc_connect.fetchall.side_effect = Exception("Fetch error")

        result = dwh_instance.execute("SELECT * FROM test_table")

        assert result == []
        assert "Failed to execute query: Fetch error" in caplog.text

    def test_set_driver_with_valid_index(self, monkeypatch):
        available_drivers = {
            "available": ["DRIVER1", "DRIVER2"],
            "supported": ["DRIVER1", "DRIVER2"],
        }
        monkeypatch.setattr(
            dwh.Dwh,
            "_Dwh__get_list_of_available_and_supported_pyodbc_drivers",
            lambda self: available_drivers,
        )

        driver_index = 1
        dwh_instance = dwh.Dwh(
            helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs(), driver_index
        )

        assert dwh_instance.driver == "DRIVER2"

    def test_get_list_of_supported_pyodbc_drivers_error(
        self, dwh_instance, monkeypatch, caplog
    ):
        monkeypatch.setattr(
            pyodbc, "drivers", mock.Mock(side_effect=pyodbc.Error("Test error"))
        )

        with pytest.raises(pyodbc.Error):
            dwh.Dwh(helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs())

    @mock.patch("pyodbc.connect")
    @mock.patch(
        "pyprediktorutilities.dwh.dwh.Dwh._Dwh__get_list_of_supported_pyodbc_drivers"
    )
    def test_get_sets_available_driver(self, mock_get_supported_drivers, mock_connect):
        mock_get_supported_drivers.return_value = ["Driver1", "Driver2"]
        mock_connect.side_effect = [pyodbc.Error, None]

        dwh_instance = dwh.Dwh(
            helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs()
        )

        assert mock_connect.call_count == 2
        assert dwh_instance.driver == "Driver2"

    @mock.patch("pyodbc.drivers")
    @mock.patch("pyprediktorutilities.dwh.dwh.Dwh._Dwh__set_driver")
    def test_get_supported_but_not_available_drivers_raises_error(
        self, mock_drivers, mock_connect
    ):
        mock_drivers.return_value = ["Driver1", "Driver2", "Driver3"]
        mock_connect.side_effect = [pyodbc.Error, pyodbc.Error, pyodbc.Error]
        dwh.Dwh(helpers.grs(), helpers.grs(), helpers.grs(), helpers.grs())

    def test_connect_success(self, dwh_instance, monkeypatch):
        mock_connection = mock.Mock()
        monkeypatch.setattr("pyodbc.connect", lambda *args, **kwargs: mock_connection)

        dwh_instance.connection = None
        dwh_instance._Dwh__connect()

        assert dwh_instance.connection is mock_connection
        assert dwh_instance.cursor is not None

    def test_connect_raises_data_error(self, dwh_instance, monkeypatch):
        def mock_connect(*args, **kwargs):
            raise pyodbc.DataError("Data error")

        monkeypatch.setattr("pyodbc.connect", mock_connect)

        dwh_instance.connection = None
        with pytest.raises(pyodbc.DataError):
            dwh_instance._Dwh__connect()

    def test_connect_raises_integrity_error(self, dwh_instance, monkeypatch):
        def mock_connect(*args, **kwargs):
            raise pyodbc.IntegrityError("Integrity error")

        monkeypatch.setattr("pyodbc.connect", mock_connect)

        dwh_instance.connection = None
        with pytest.raises(pyodbc.IntegrityError):
            dwh_instance._Dwh__connect()

    def test_connect_raises_programming_error(self, dwh_instance, monkeypatch):
        def mock_connect(*args, **kwargs):
            raise pyodbc.ProgrammingError("Programming error")

        monkeypatch.setattr("pyodbc.connect", mock_connect)

        dwh_instance.connection = None
        with pytest.raises(pyodbc.ProgrammingError):
            dwh_instance._Dwh__connect()

    def test_connect_raises_not_supported_error(self, dwh_instance, monkeypatch):
        def mock_connect(*args, **kwargs):
            raise pyodbc.NotSupportedError("Not supported error")

        monkeypatch.setattr("pyodbc.connect", mock_connect)

        dwh_instance.connection = None
        with pytest.raises(pyodbc.NotSupportedError):
            dwh_instance._Dwh__connect()

    def test_connect_retries_on_operational_error(self, dwh_instance, monkeypatch):
        attempt_count = 0

        def mock_connect(*args, **kwargs):
            nonlocal attempt_count
            attempt_count += 1
            if attempt_count < 3:
                raise pyodbc.OperationalError("Operational error")
            return mock.Mock()

        dwh_instance.connection_attempts = 3
        dwh_instance.connection = None

        with mock.patch("pyodbc.connect", side_effect=mock_connect):
            dwh_instance._Dwh__connect()

        assert attempt_count == 3, "Should attempt three connections before succeeding"
        assert (
            dwh_instance.connection is not None
        ), "Connection should be established after retries"

    def test_connect_raises_after_max_attempts_on_operational_error(
        self, dwh_instance, monkeypatch
    ):
        def mock_connect(*args, **kwargs):
            raise pyodbc.OperationalError("Operational error")

        dwh_instance.connection = None
        dwh_instance.connection_attempts = 3
        with mock.patch("pyodbc.connect", side_effect=mock_connect):
            with pytest.raises(pyodbc.Error, match="Failed to connect to the database"):
                dwh_instance._Dwh__connect()

    def test_connect_retries_on_database_error(self, dwh_instance, monkeypatch):
        attempt_count = 0

        def mock_connect(*args, **kwargs):
            nonlocal attempt_count
            attempt_count += 1
            if attempt_count < 3:
                raise pyodbc.DatabaseError("Database error")
            return mock.Mock()

        dwh_instance.connection_attempts = 3
        dwh_instance.connection = None

        with mock.patch("pyodbc.connect", side_effect=mock_connect):
            dwh_instance._Dwh__connect()

        assert attempt_count == 3, "Should attempt three connections before succeeding"

    def test_connect_raises_after_max_attempts_on_database_error(
        self, dwh_instance, monkeypatch
    ):
        def mock_connect(*args, **kwargs):
            raise pyodbc.DatabaseError("Database error")

        dwh_instance.connection = None
        dwh_instance.connection_attempts = 3
        with mock.patch("pyodbc.connect", side_effect=mock_connect):
            with pytest.raises(pyodbc.Error, match="Failed to connect to the database"):
                dwh_instance._Dwh__connect()

    def test_connect_retries_on_generic_error(self, dwh_instance, monkeypatch):
        attempt_count = 0

        def mock_connect(*args, **kwargs):
            nonlocal attempt_count
            attempt_count += 1
            if attempt_count < 3:
                raise pyodbc.Error("Generic error")
            return mock.Mock()

        dwh_instance.connection_attempts = 3
        dwh_instance.connection = None

        with mock.patch("pyodbc.connect", side_effect=mock_connect):
            dwh_instance._Dwh__connect()

        assert attempt_count == 3, "Should attempt three connections before succeeding"

    def test_connect_raises_after_max_attempts_on_generic_error(
        self, dwh_instance, monkeypatch
    ):
        def mock_connect(*args, **kwargs):
            raise pyodbc.Error("Generic error")

        dwh_instance.connection = None
        dwh_instance.connection_attempts = 3
        with mock.patch("pyodbc.connect", side_effect=mock_connect):
            with pytest.raises(pyodbc.Error, match="Failed to connect to the database"):
                dwh_instance._Dwh__connect()

    def test_connect_raises_error_when_connection_is_none(
        self, dwh_instance, monkeypatch
    ):

        def mock_connect(*args, **kwargs):
            return None

        monkeypatch.setattr("pyodbc.connect", mock_connect)

        dwh_instance.connection_attempts = 3
        dwh_instance.connection = None

        with pytest.raises(pyodbc.Error, match="Failed to connect to the database"):
            dwh_instance._Dwh__connect()

    def test_are_connection_attempts_reached(self, dwh_instance, caplog):
        assert not dwh_instance._Dwh__are_connection_attempts_reached(1)
        assert "Retrying connection..." in caplog.text

        assert dwh_instance._Dwh__are_connection_attempts_reached(3)
        assert "Failed to connect to the DataWarehouse after 3 attempts." in caplog.text

    def test_disconnect(self, dwh_instance):
        mock_connection = mock.Mock()
        dwh_instance.connection = mock_connection
        dwh_instance.cursor = mock.Mock()

        dwh_instance._Dwh__disconnect()

        assert mock_connection.close.called
        assert dwh_instance.connection is None
        assert dwh_instance.cursor is None

    def test_disconnect_without_connection(self, dwh_instance):
        dwh_instance.connection = None
        dwh_instance.cursor = None

        dwh_instance._Dwh__disconnect()

    def test_commit(self, dwh_instance):
        mock_connection = mock.Mock()
        dwh_instance.connection = mock_connection
        dwh_instance._Dwh__commit()
        mock_connection.commit.assert_called_once()



import random
import string

import pyodbc


def mock_pyodbc_connection_throws_error_not_tolerant_to_attempts(connection_string):
    raise pyodbc.DataError("Error code", "Error message")


def mock_pyodbc_connection_throws_error_tolerant_to_attempts(connection_string):
    raise pyodbc.DatabaseError("Error code", "Error message")


def grs():
    """Generate a random string."""
    return "".join(random.choices(string.ascii_uppercase + string.digits, k=10))
